package secrets_manager

import (
	"context"
	"encoding/base64"
	"encoding/json"
	"errors"
	"os"

	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/logs"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/struct_utils"

	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/config"
	"github.com/aws/aws-sdk-go-v2/credentials"
	"github.com/aws/aws-sdk-go-v2/service/secretsmanager"
	"github.com/aws/aws-secretsmanager-caching-go/secretcache"
	"github.com/aws/smithy-go"
)

type DatabaseCredentials struct {
	Username           string `json:"username"`
	Password           string `json:"password"`
	Engine             string `json:"engine"`
	Host               string `json:"host"`
	Port               int    `json:"port"`
	InstanceIdentifier string `json:"dbInstanceIdentifier"`
	ReadOnlyHost       string `json:"aurora_read_only_host"`
}

type S3UploadCredentials struct {
	AccessKeyID string `json:"accessKeyID"`
	SecretKey   string `json:"secretKey"`
}

var (
	secretCache, _      = secretcache.New()
	secretManagerRegion = "af-south-1"
)

var secretManagerClient *secretsmanager.Client

func GetDatabaseCredentials(secretID string, isDebug bool) (DatabaseCredentials, error) {
	secret, _ := GetSecret(secretID, isDebug)
	var credentials DatabaseCredentials
	err := struct_utils.UnmarshalJSON([]byte(secret), &credentials)
	if err != nil {
		return DatabaseCredentials{}, err
	}
	return credentials, nil
}

func GetS3UploadCredentialsProvider(secretID string, isDebug bool) (credentials.StaticCredentialsProvider, error) {
	var secretValue S3UploadCredentials
	if secretID != "" {
		secret, _ := GetSecret(secretID, isDebug)
		err := struct_utils.UnmarshalJSON([]byte(secret), &secretValue)
		if err != nil {
			return credentials.StaticCredentialsProvider{}, err
		}
	}

	// Attempt to get credentials from the environment variables if not set in the secret
	if secretValue.AccessKeyID == "" && secretValue.SecretKey == "" {
		secretValue.AccessKeyID = os.Getenv("AWS_ACCESS_KEY")
		secretValue.SecretKey = os.Getenv("AWS_SECRET_KEY")
	}

	if secretValue.AccessKeyID == "" || secretValue.SecretKey == "" {
		return credentials.StaticCredentialsProvider{}, nil
	}

	return credentials.NewStaticCredentialsProvider(secretValue.AccessKeyID, secretValue.SecretKey, ""), nil
}

func instantiateSecretManagerClient(isDebug bool) (err error) {
	// If a client already exists, use it
	if secretManagerClient != nil {
		return nil
	}

	cfg, err := config.LoadDefaultConfig(context.TODO())
	if err != nil {
		return err
	}

	// Get local config
	if isDebug && os.Getenv("ENVIRONMENT") != "" {
		cfg, err = config.LoadDefaultConfig(context.TODO(),
			config.WithRegion(secretManagerRegion),
		)
		if err != nil {
			return err
		}
	}

	// Create a Secrets Manager client
	secretManagerClient = secretsmanager.NewFromConfig(cfg)

	return nil
}

// logError Logs any errors returned by the Secrets Manager client
func logError(err error) {
	var apiErr smithy.APIError
	if errors.As(err, &apiErr) {
		logs.Info(apiErr.ErrorCode()+" %s", apiErr.ErrorMessage())
	} else {
		logs.Info(err.Error())
	}
}

func GetSecret(secretID string, isDebug bool) (string, string) {
	// Check if we have the secret in cache
	cachedSecret, _ := secretCache.GetSecretString(secretID)
	if cachedSecret != "" {
		return cachedSecret, ""
	}

	// Create a Secrets Manager client
	err := instantiateSecretManagerClient(isDebug)
	if err != nil {
		logs.Info("Could not create client: %+v", err)
		return "", ""
	}

	// Create a secret
	input := &secretsmanager.GetSecretValueInput{
		SecretId:     aws.String(string(secretID)),
		VersionStage: aws.String("AWSCURRENT"), // VersionStage defaults to AWSCURRENT if unspecified
	}

	result, err := secretManagerClient.GetSecretValue(context.TODO(), input)
	if err != nil {
		logError(err)
		return "", ""
	}

	// Decrypts secret using the associated KMS CMK.
	// Depending on whether the secret is a string or binary, one of these fields will be populated.
	var secretString, decodedBinarySecret string
	if result.SecretString != nil {
		secretString = *result.SecretString
	} else {
		decodedBinarySecretBytes := make([]byte, base64.StdEncoding.DecodedLen(len(result.SecretBinary)))
		length, err := base64.StdEncoding.Decode(decodedBinarySecretBytes, result.SecretBinary)
		if err != nil {
			logs.Info("Base64 Decode Error: %+v", err)
			return "", ""
		}
		decodedBinarySecret = string(decodedBinarySecretBytes[:length])
	}

	return secretString, decodedBinarySecret
}

// CreateSecret Creates a JSON marshaled "string secret" (can be expanded to cater for binary secrets should the need arise)
func CreateSecret(secretID string, secret any, isDebug bool) (awsSecretID string, err error) {
	// Create a Secrets Manager client
	err = instantiateSecretManagerClient(isDebug)
	if err != nil {
		logs.Info("Could not create client: %+v", err)
		return "", err
	}

	// Create the secret - marshaling "any" into a JSON string
	secretStr, err := json.Marshal(secret)
	if err != nil {
		logs.Info("Could not marshal secret: %+v", err)
		return "", err
	}
	input := &secretsmanager.CreateSecretInput{
		Name:         aws.String(secretID),
		SecretString: aws.String(string(secretStr)),
	}

	result, err := secretManagerClient.CreateSecret(context.TODO(), input)
	if err != nil {
		logError(err)
		return "", err
	}

	return aws.ToString(result.Name), nil
}

func DeleteSecret(secretID string, forceWithoutRecovery bool, isDebug bool) error {
	// Create a Secrets Manager client
	err := instantiateSecretManagerClient(isDebug)
	if err != nil {
		logs.Info("Could not create client: %+v", err)
		return err
	}

	// Delete the secret
	input := &secretsmanager.DeleteSecretInput{
		SecretId:                   aws.String(secretID),
		ForceDeleteWithoutRecovery: aws.Bool(forceWithoutRecovery),
	}

	_, err = secretManagerClient.DeleteSecret(context.TODO(), input)
	if err != nil {
		logError(err)
		return err
	}

	return nil
}

// UpdateSecret Updates an exising secret
func UpdateSecret(secretID string, secret any, isDebug bool) error {
	// Create a Secrets Manager client
	err := instantiateSecretManagerClient(isDebug)
	if err != nil {
		logs.Info("Could not create client: %+v", err)
		return err
	}

	// Create the secret - marshaling "any" into a JSON string
	secretStr, err := json.Marshal(secret)
	if err != nil {
		logs.Info("Could not marshal secret: %+v", err)
		return err
	}
	input := &secretsmanager.UpdateSecretInput{
		SecretId:     aws.String(secretID),
		SecretString: aws.String(string(secretStr)),
	}

	_, err = secretManagerClient.UpdateSecret(context.TODO(), input)
	if err != nil {
		logError(err)
		return err
	}

	return nil
}
