package websocket_utils

import (
	"context"
	"fmt"
	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/config"
	"github.com/aws/aws-sdk-go-v2/service/apigatewaymanagementapi"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/logs"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/utils"
	"os"
)

var (
	sessions = map[string]*APIGateWayClientWithHelpers{}
)

type APIGateWayClientWithHelpers struct {
	APIGatewayClient *apigatewaymanagementapi.Client
	Available        bool
}

func GetClient(region ...string) *APIGateWayClientWithHelpers {
	domainName := os.Getenv("WEBSOCKET_DOMAIN_NAME")
	if domainName == "" {
		logs.ErrorMsg("WEBSOCKET_DOMAIN_NAME env variable is not set")
		return &APIGateWayClientWithHelpers{}
	}
	s3Region := os.Getenv("AWS_REGION")

	// Set custom region
	if region != nil && len(region) > 0 {
		s3Region = region[0]
	}

	// Check if session exists for region, if it does return it
	if apiGatewaySession, ok := sessions[s3Region]; ok {
		return apiGatewaySession
	}

	// Setup session
	cfg, err := config.LoadDefaultConfig(context.TODO(),
		config.WithRegion(s3Region),
	)
	if err != nil {
		logs.ErrorMsg("failed to load AWS config for websocket client")
		return &APIGateWayClientWithHelpers{}
	}

	apiGatewaySession := NewClient(cfg, domainName)
	sessions[s3Region] = apiGatewaySession
	return apiGatewaySession
}

func NewClient(config aws.Config, domainName string) *APIGateWayClientWithHelpers {
	return &APIGateWayClientWithHelpers{
		APIGatewayClient: apigatewaymanagementapi.NewFromConfig(config, func(o *apigatewaymanagementapi.Options) {
			o.BaseEndpoint = utils.ValueToPointer(fmt.Sprintf("https://%s/default", domainName))
		}),
		Available: true,
	}
}

func (a APIGateWayClientWithHelpers) PostToConnectionID(data []byte, connectionID string) error {
	if !a.Available {
		return nil
	}

	_, err := a.APIGatewayClient.PostToConnection(context.TODO(), &apigatewaymanagementapi.PostToConnectionInput{
		ConnectionId: &connectionID,
		Data:         data,
	})
	return err
}

func (a APIGateWayClientWithHelpers) PostToConnectionIDs(data []byte, connectionIDs []string) {
	if !a.Available {
		return
	}

	for _, connectionID := range connectionIDs {
		_, err := a.APIGatewayClient.PostToConnection(context.TODO(), &apigatewaymanagementapi.PostToConnectionInput{
			ConnectionId: &connectionID,
			Data:         data,
		})
		if err != nil {
			logs.Info("failed to send data to connection ID: %s", connectionID)
		}
	}
}
